import itertools
import logging as log
from typing import Optional, Union, List, Dict, Sequence, Iterable, Collection, Callable

import torch
import torch.nn as nn
import tinycudann as tcnn

from ops.interpolation import grid_sample_wrapper
from raymarching.spatial_distortions import SpatialDistortion
import pdb


def HSV2RGB(H:torch.Tensor,S:torch.Tensor,V:torch.Tensor):
    # batchsize, 1
    #https://www.had2know.org/technology/hsv-rgb-conversion-formula-calculator.html

    M = 255*V 
    m = M*(1-S)
    z = (M-m) * (1-torch.abs((H/60)%2 -1 ))
    #6 cases
    rgb = torch.where(H < 60, torch.cat([M, z, torch.zeros_like(M)], dim=1),
                      torch.where(H < 120, torch.cat([z, M, torch.zeros_like(M)], dim=1),
                                  torch.where(H < 180, torch.cat([torch.zeros_like(M), M, z], dim=1),
                                              torch.where( H < 240, torch.cat([torch.zeros_like(M), z, M], dim=1),
                                                          torch.where(H < 300, torch.cat([z, torch.zeros_like(M), M], dim=1),
                                                                      torch.cat([M, torch.zeros_like(M), z], dim=1))))))

    rgb = rgb + m
    rgb = rgb/255
    return rgb

def get_normalized_directions(directions):
    """SH encoding must be in the range [0, 1]

    Args:
        directions: batch of directions
    """
    return (directions + 1.0) / 2.0


def normalize_aabb(pts, aabb):
    return (pts - aabb[0]) * (2.0 / (aabb[1] - aabb[0])) - 1.0


def init_grid_param(
        grid_nd: int,
        in_dim: int,
        out_dim: int,
        reso: Sequence[int],
        a: float = 0.1,
        b: float = 0.5):
    assert in_dim == len(reso), "Resolution must have same number of elements as input-dimension"
    has_time_planes = in_dim == 4
    assert grid_nd <= in_dim
    coo_combs = list(itertools.combinations(range(in_dim), grid_nd)) 
    grid_coefs = nn.ParameterList()
    for ci, coo_comb in enumerate(coo_combs):
        new_grid_coef = nn.Parameter(torch.empty(
            [1, out_dim] + [reso[cc] for cc in coo_comb[::-1]]
        )) 
        if has_time_planes and 3 in coo_comb:  # Initialize time planes to 1
            nn.init.ones_(new_grid_coef)
        else:
            nn.init.uniform_(new_grid_coef, a=a, b=b)
        grid_coefs.append(new_grid_coef)

    return grid_coefs


def interpolate_ms_features(pts: torch.Tensor,
                            ms_grids: Collection[Iterable[nn.Module]],
                            grid_dimensions: int,
                            concat_features: bool,
                            num_levels: Optional[int],
                            ) -> torch.Tensor:
    coo_combs = list(itertools.combinations(
        range(pts.shape[-1]), grid_dimensions)
    )
    if num_levels is None:
        num_levels = len(ms_grids)
    multi_scale_interp = [] if concat_features else 0.
    grid: nn.ParameterList
    try:
        for scale_id, grid in enumerate(ms_grids[:num_levels]):
            interp_space = 1.
            for ci, coo_comb in enumerate(coo_combs):
                # interpolate in plane
                feature_dim = grid[ci].shape[1]  # shape of grid[ci]: 1, out_dim, *reso
                try:

                    interp_out_plane = (
                        grid_sample_wrapper(grid[ci], pts[..., coo_comb])
                        .view(-1, feature_dim)
                    )
                    # compute product over planes
                    interp_space = interp_space * interp_out_plane
                except:
                    pdb.set_trace

            # combine over scales
            if concat_features:
                multi_scale_interp.append(interp_space)
            else:
                multi_scale_interp = multi_scale_interp + interp_space

        if concat_features:
            multi_scale_interp = torch.cat(multi_scale_interp, dim=-1)
    except:
        pdb.set_trace
    return multi_scale_interp

class Lplane_field(nn.Module):
    def __init__(
        self,
        aabb,
        grid_config: Union[str, List[Dict]],
        concat_features_across_scales: bool,
        multiscale_res: Optional[Sequence[int]],
        spatial_distortion: Optional[SpatialDistortion],
        density_activation: Callable,
        linear_decoder: bool,
        linear_decoder_layers: Optional[int],
        k:Optional[int],
        num_frames: int,
        use_motion: bool,
        hues:Optional[torch.Tensor],
        solid_threshold: float,
    ) -> None:
        super().__init__()
        self.use_motion = use_motion
        self.solid_threshold = solid_threshold
        self.time_interval = 2/num_frames
        self.aabb = nn.Parameter(aabb, requires_grad=False)
        self.spatial_distortion = spatial_distortion
        self.grid_config = grid_config
        self.num_frames = num_frames
        self.multiscale_res_multipliers: List[int] = multiscale_res or [1]
        self.concat_features = concat_features_across_scales
        self.density_activation = density_activation
        self.linear_decoder = linear_decoder
        if self.use_motion:
            self.motion_dim = 3
        else:
            self.motion_dim = 0
        # 1. Init planes
        self.grids = nn.ModuleList()
        self.feature_dim = 0
        for res in self.multiscale_res_multipliers:
            # initialize coordinate grid
            config = self.grid_config[0].copy()
            # Resolution fix: multi-res only on spatial planes
            config["resolution"] = [
                r * res for r in config["resolution"][:3]
            ] + config["resolution"][3:]
            gp = init_grid_param(
                grid_nd=config["grid_dimensions"],
                in_dim=config["input_coordinate_dim"],
                out_dim=config["output_coordinate_dim"],
                reso=config["resolution"],
            )
            # shape[1] is out-dim - Concatenate over feature len for each scale
            if self.concat_features:
                self.feature_dim += gp[-1].shape[1]
            else:
                self.feature_dim = gp[-1].shape[1]
            self.grids.append(gp)
        log.info(f"Initialized model grids: {self.grids}")
        self.num_frames = num_frames
        if hues is not None:
            self.original = nn.Parameter(360*hues,requires_grad=False)
            self.hues = nn.Parameter(360*hues,requires_grad=True) 
        else:
            self.original = nn.Parameter(torch.zeros(num_frames,k),requires_grad=False)
            self.hues = nn.Parameter(torch.zeros(num_frames,k),requires_grad=True) 

        # 3. Init decoder params
        self.direction_encoder = tcnn.Encoding(
            n_input_dims=3,
            encoding_config={
                "otype": "SphericalHarmonics",
                "degree": 4,
            },
        )

        # 3. Init decoder network
        """
        if self.linear_decoder:
            assert linear_decoder_layers is not None
            # The NN learns a basis that is used instead of spherical harmonics
            # Input is an encoded view direction, output is weights for
            # combining the color features into RGB
            # This architecture is based on instant-NGP
            self.color_basis = tcnn.Network(
                n_input_dims=3 ,#self.direction_encoder.n_output_dims,
                n_output_dims=3 * self.feature_dim,
                network_config={
                    "otype": "FullyFusedMLP",
                    "activation": "ReLU",
                    "output_activation": "None",
                    "n_neurons": 128,
                    "n_hidden_layers": linear_decoder_layers,
                },
            )
            # sigma_net just does a linear transformation on the features to get density
            self.sigma_net = tcnn.Network(
                n_input_dims=self.feature_dim,
                n_output_dims=1,
                network_config={
                    "otype": "CutlassMLP",
                    "activation": "None",
                    "output_activation": "None",
                    "n_neurons": 128,
                    "n_hidden_layers": 0,
                },
            )
        else:
        """
        self.geo_feat_dim = 15
        self.sigma_net = tcnn.Network(
            n_input_dims=self.feature_dim,
            n_output_dims=self.geo_feat_dim + self.motion_dim + 1,
            network_config={
                "otype": "FullyFusedMLP",
                "activation": "ReLU",
                "output_activation": "None",
                "n_neurons": 64,
                "n_hidden_layers": 1,
            },
        )
        self.in_dim_color = (
                self.direction_encoder.n_output_dims
                + self.geo_feat_dim
        )
        self.SV_net = tcnn.Network(
            n_input_dims=self.in_dim_color,
            n_output_dims=2, #output s,v
            network_config={
                "otype": "FullyFusedMLP",
                "activation": "ReLU",
                "output_activation": "None",
                "n_neurons": 64,
                "n_hidden_layers": 2,
            },
            
        )
        self.Hue_select = tcnn.Network(
            n_input_dims=self.in_dim_color,
            n_output_dims=k, #output s,v
            network_config={
                "otype": "FullyFusedMLP",
                "activation": "ReLU",
                "output_activation": "None",
                "n_neurons": 64,
                "n_hidden_layers": 2,
            },
            
        )
    def get_features_w_latent_code(self, features, timestamps, n_rays, n_samples):
        
        timestamps = timestamps.reshape(n_rays*n_samples)
        _idx = ((timestamps+1)/2)*(self.num_frames-1)
        _idx = _idx.view(-1,1)
        idx = _idx.to(torch.int32)
        dff = _idx-idx
        
        interp = (torch.index_select(self.latent_code, 0, idx[0])*(1-dff) +
                torch.index_select(self.latent_code, 0, idx[0]+1)*(dff))
        
        features = features * interp
        
        return features

    def get_hues(self,  timestamps, n_rays, n_samples):
        timestamps = timestamps.reshape(n_rays*n_samples)
        _idx = ((timestamps+1)/2)*(self.num_frames)
        _idx = _idx.view(-1,1)
        idx = _idx.to(torch.int32)
        ceil_idx = torch.ceil(_idx).to(torch.int32)
        dff = _idx-idx
    
        interp = (torch.index_select(self.hues, 0, idx[:,0])*(1-dff) + torch.index_select(self.hues, 0, ceil_idx[:,0])*(dff))
        
        return interp

    def decouple(self,
                pts: torch.Tensor,
                directions: torch.Tensor,
                light_time: Optional[torch.Tensor] = None,
                human_time: Optional[torch.Tensor] = None):
        #pdb.set_trace()
        light_time_density, light_features, light_hues = self.get_density(pts,light_time)
        human_time_density, human_features, human_hues = self.get_density(pts,human_time)

        n_rays, n_samples = pts.shape[:2]

        directions = directions.view(-1, 1, 3).expand(pts.shape).reshape(-1, 3)
       # hues = hues.view(-1,1, hues.shape[-1]).expand(pts.shape[0],pts.shape[1],hues.shape[-1]).reshape(-1, hues.shape[-1])
        if not self.linear_decoder:
            directions = get_normalized_directions(directions)
            encoded_directions = self.direction_encoder(directions)

        if self.linear_decoder:
            color_human_features = [human_features]
            color_light_features = [light_features]

        else:
            color_human_features = [encoded_directions, human_features.view(-1, self.geo_feat_dim)]
            color_light_features = [encoded_directions, light_features.view(-1, self.geo_feat_dim)]

        color_human_features = torch.cat(color_human_features, dim=-1)
        color_light_features = torch.cat(color_light_features, dim=-1)

        S, V = torch.split(torch.sigmoid(self.SV_net(color_light_features)).to(directions),[1,1],dim=-1)#.view(n_rays, n_samples, 3)
        H =  torch.sum(light_hues*torch.softmax(self.Hue_select(color_light_features),dim=-1),dim=-1,keepdim=True).to(directions)
        rgb = HSV2RGB(H,S,V).view(n_rays,n_samples,3)

        return {"l_color": rgb.nan_to_num(1e-6), "density_il": human_time_density.nan_to_num(1e-6), 'hues':light_hues.view(n_rays,n_samples,light_hues.shape[-1]).nan_to_num(1e-6)}
    def custom_hue(self,
                pts: torch.Tensor,
                directions: torch.Tensor,
                custom_Hue: torch.Tensor,
                light_time: Optional[torch.Tensor] = None,
                human_time: Optional[torch.Tensor] = None):
        #pdb.set_trace()
        light_time_density, light_features, light_hues = self.get_density(pts,light_time)
        human_time_density, human_features, human_hues = self.get_density(pts,human_time)

        n_rays, n_samples = pts.shape[:2]

        directions = directions.view(-1, 1, 3).expand(pts.shape).reshape(-1, 3)
       # hues = hues.view(-1,1, hues.shape[-1]).expand(pts.shape[0],pts.shape[1],hues.shape[-1]).reshape(-1, hues.shape[-1])
        if not self.linear_decoder:
            directions = get_normalized_directions(directions)
            encoded_directions = self.direction_encoder(directions)

        if self.linear_decoder:
            color_human_features = [human_features]
            color_light_features = [light_features]

        else:
            color_human_features = [encoded_directions, human_features.view(-1, self.geo_feat_dim)]
            color_light_features = [encoded_directions, light_features.view(-1, self.geo_feat_dim)]

        color_human_features = torch.cat(color_human_features, dim=-1)
        color_light_features = torch.cat(color_light_features, dim=-1)

        S, V = torch.split(torch.sigmoid(self.SV_net(color_light_features)).to(directions),[1,1],dim=-1)#.view(n_rays, n_samples, 3)
        H =  custom_Hue* torch.ones_like(S,device=S.device)
        rgb = HSV2RGB(H,S,V).view(n_rays,n_samples,3)

        return {"l_color": rgb.nan_to_num(1e-6), "density_il": human_time_density.nan_to_num(1e-6), 'hues':light_hues.view(n_rays,n_samples,light_hues.shape[-1]).nan_to_num(1e-6)}





    def get_density(self, pts: torch.Tensor, timestamps: Optional[torch.Tensor] = None):
        """Computes and returns the densities."""

        if self.spatial_distortion is not None:
            pts = self.spatial_distortion(pts)
            pts = pts / 2  # from [-2, 2] to [-1, 1]
        else:
            pts = normalize_aabb(pts, self.aabb)
            
        n_rays, n_samples = pts.shape[:2]
        if timestamps is not None: # not none [4096]
            timestamps = timestamps[:, None].expand(-1, n_samples)[..., None]  # [n_rays, n_samples, 1]
            pts = torch.cat((pts, timestamps), dim=-1)  # [n_rays, n_samples, 4]
        pts = pts.reshape(-1, pts.shape[-1])
        features = interpolate_ms_features(
            pts, ms_grids=self.grids,  # noqa
            grid_dimensions=self.grid_config[0]["grid_dimensions"],
            concat_features=self.concat_features, num_levels=None)
        hues = self.get_hues(timestamps, n_rays, n_samples)
        #hues[hues>360] = hues[hues>360] % 360
        #    features = self.get_features_w_latent_code(features, timestamps, n_rays, n_samples)
        if len(features) < 1:
            features = torch.zeros((0, 1)).to(features.device)
        if self.linear_decoder:
            density_before_activation = torch.split(
                self.sigma_net(features), [1], dim=-1)  # [batch, 1]
        else:
            features = self.sigma_net(features)
            features, density_before_activation = torch.split(
                features, [self.geo_feat_dim, 1], dim=-1) #[196608,15]

        density = self.density_activation(
            density_before_activation.to(pts)
        ).view(n_rays, n_samples, 1) #density: [4096,48,1]

        return density, features, hues
    


    def forward(self,
                pts: torch.Tensor,
                directions: torch.Tensor,
                timestamps: Optional[torch.Tensor] = None):
        #pdb.set_trace()
        density,features, hues = self.get_density(pts, timestamps)
        n_rays, n_samples = pts.shape[:2]

        directions = directions.view(-1, 1, 3).expand(pts.shape).reshape(-1, 3)
       # hues = hues.view(-1,1, hues.shape[-1]).expand(pts.shape[0],pts.shape[1],hues.shape[-1]).reshape(-1, hues.shape[-1])
        if not self.linear_decoder:
            directions = get_normalized_directions(directions)
            encoded_directions = self.direction_encoder(directions)

        if self.linear_decoder:
            color_features = [features]
        else:
            color_features = [encoded_directions, features.view(-1, self.geo_feat_dim)]
            

        color_features = torch.cat(color_features, dim=-1)
        S, V = torch.split(torch.sigmoid(self.SV_net(color_features)).to(directions),[1,1],dim=-1)#.view(n_rays, n_samples, 3)
        H =  torch.sum(hues*torch.softmax(self.Hue_select(color_features),dim=-1),dim=-1,keepdim=True).to(directions)
        rgb = HSV2RGB(H,S,V).view(n_rays,n_samples,3)

        return {"l_color": rgb.nan_to_num(1e-6), "density_il": density.nan_to_num(1e-6), 'hues':hues.view(n_rays,n_samples,hues.shape[-1]).nan_to_num(1e-6)}

    def get_params(self):
        field_params = {k: v for k, v in self.grids.named_parameters(prefix="grids")}
        nn_params = [
            self.sigma_net.named_parameters(prefix="sigma_net"),
            self.direction_encoder.named_parameters(prefix="direction_encoder"),
        ]
        nn_params.append(self.Hue_select.named_parameters(prefix="Hue_select"))
        nn_params.append(self.SV_net.named_parameters(prefix="SV_net"))
        hues = [self.hues]
        nn_params = {k: v for plist in nn_params for k, v in plist}
        other_params = {k: v for k, v in self.named_parameters() if (
            k not in nn_params.keys() and k not in field_params.keys() and k != "hues"
        )}
        return {
            "nn": list(nn_params.values()),
            "field": list(field_params.values()),
            "hues" : hues,
            "other": list(other_params.values()),
            
        }
        

         